-- Biodiversity by
-- dave@beermex.com

audio_handler={}
local mt = {}

function audio_handler:new( file, looping )
	if looping==nil then looping=false end
    return( setmetatable({ file = file, looping = looping, sources = {} }, mt ))
end

mt.__index = audio_handler

function audio_handler:play( index, pitch )
	if self.file == "" then return end

	if pitch==nil then pitch=1 end

	local sources = self.sources
    if index == nil then
	    for s_index=1,#sources do
			assert( sources[s_index]~=nil )
		    if sources[s_index]:isStopped() then index=s_index;break; end
		end
		if index==nil then index=1 end
	end
	assert(index ~= nil )
	if sources[index]==nil then
		sources[index] = love.audio.newSource( self.file, 'streaming' )
		sources[index]:setLooping(self.looping)
	end

	assert( sources[index] ~= nil )
	if not sources[index]:isStopped() then sources[index]:stop() end
	assert( sources[index]:isStopped() )
	sources[index]:setVolume(1)
	sources[index]:setPitch(pitch)
	sources[index]:play()
	return index
end

function audio_handler:stop( index )
	if self.file == "" then return end

	local sources = self.sources
    assert( index ~= nil )
	if sources[index] ~= nil and not sources[index]:isStopped() then
		sources[index]:setVolume(0.01)
		sources[index]:stop()
	end
end

function audio_handler:stopAll()
	if self.file == "" then return end

	local sources = self.sources
    for index=1,#sources do
		if sources[index] ~= nil and not sources[index]:isStopped() then
			sources[index]:setVolume(0.01)
			sources[index]:stop()
		end
	end
end

function audio_handler:pauseAll()
	if self.file == "" then return end

	local sources = self.sources
    for index=1,#sources do
		if sources[index] ~= nil and not sources[index]:isStopped() then
			sources[index]:pause()
		end
	end
end

function audio_handler:resumeAll()
	if self.file == "" then return end

	local sources = self.sources
    for index=1,#sources do
		if sources[index] ~= nil and sources[index]:isPaused() then
			sources[index]:resume()
		end
	end
end
