-- Biodiversity by
-- dave@beermex.com

flowers = {}
flowergrid = {}
flowergridfilled = 0

gridheight = 8
gridwidth = 8

function flowers_setup()
	local x
	local y
	for x=0,gridwidth-1 do
		flowergrid[x]={}
		for y=0,gridheight-1 do
			flowergrid[x][y]=-1
		end
	end
	flowergridfilled = 0
end

function flowers_winning_dude()
	-- mark the matched guys as winning dudes.
	for i=1,#flowers do
		if flowers[i].g_match then
			flowers[i].winning_dude=true
		else
			flowers[i].winning_dude=false
			flowers[i].pollinated=false
			if flowers[i].state == 'pollinated' or flowers[i].state == 'fruitgrowing' then
				flowers[i].state = 'living'
			end
		end
	end
end

function addflower( gridx, gridy, sourcegridx, sourcegridy, sourceox, sourceoy, genes1, genes2, mrate )

	-- x,y are in grid coordinates;  the position of the flower.  we need these to be
	-- in gridcoords to ensure the flowergrid can be looked up like a matrix.

	-- sourcegridx, sourcegridy == target position of fruit  (x,y are dynamic until x,y == sourcex,sourcey)
	-- This is what we set when we create a new plant (i.e. create a fruit).  X,Y are NOT where the fruit will
	-- be planted;  sourcegridx,sourcegridy ARE.
	-- these are in grid coordinates

	-- sourceox, sourceoy == initial location of fruit (i.e. on the fruit-bearing flower)
	-- these are in world coordinates

	if gridx==nil and gridy==nil then

		--grid is full!
		if flowergridfilled == gridwidth*gridheight then
			return nil
		end

		local ok=false
		while not ok do
			gridx = math.random(gridwidth)-1
			gridy = math.random(gridheight)-1
			assert( gridx <= gridwidth-1 and gridx >= 0 )
			assert( flowergrid[gridx] ~= nil )
			assert( gridy <= gridheight-1 and gridy >= 0 )
			assert( flowergrid[gridx][gridy]~=nil)
			if flowergrid[gridx][gridy] == -1 then
				ok = true
			end
		end

	end

	assert( gridx >= 0 and gridx <= (gridwidth-1) --[[, "x="..x.." #flowergrid="..#flowergrid]] )
	assert( flowergrid[gridx] ~= nil )
	assert( gridy >= 0 and gridy <= (gridheight-1) --[[, "y="..x.." #flowergrid[x]="..#(flowergrid[x])]] )

	assert( flowergrid[gridx][gridy]==-1 )

	newflowerid = #flowers+1

	flowergrid[gridx][gridy]= newflowerid
	flowergridfilled = flowergridfilled + 1

	if genes1 ~= nil and genes2==nil then
		genes = genes1
	else
		if genes1==nil and genes2== nil then
			colorgene = Gene:new('Color',{2,2,2}, 'count', {math.random(2^(2*2))-1,math.random(2^(2*2))-1,math.random(2^(2*2))-1} )
			heightgene = Gene:new('Height', {2}, 'value', {math.random(2^(2*2))-1} )
			poisongene = Gene:new('Poison',{6}, 'count', {math.random(2^(6*2))-1})
			numpetals = Gene:new('NumPetals',{3}, 'value', {math.random(2^(3*2))-1} )
			attractiveness = Gene:new('Attractiveness', {4}, 'count', {math.random( 2^(4*2))-1} )
		else
			assert( genes1 ~= nil and genes2 ~= nil )
			colorgene = genes1.colorgene:copy()
			colorgene:breed(genes2.colorgene,mrate)

			heightgene = genes1.heightgene:copy()
			heightgene:breed(genes2.heightgene,mrate)

			poisongene = genes1.poisongene:copy()
			poisongene:breed(genes2.poisongene,mrate)

			numpetals = genes1.numpetals:copy()
			numpetals:breed(genes2.numpetals,mrate)

			attractiveness = genes1.attractiveness:copy()
			attractiveness:breed(genes2.attractiveness,mrate)
		end
		genes = {colorgene=colorgene, heightgene=heightgene, poisongene=poisongene, numpetals=numpetals, attractiveness=attractiveness}
	end

	local heightgenevalue = heightgene:getValueAsAlleleValues()[1]
	local heightgenev = 1+(heightgenevalue.value/heightgenevalue.maxvalue) -- number between 1 and 2
	assert( heightgenev >= 1 and heightgenev <= 2 )
	local flowersize = FLOWERSIZEMIN + ((heightgenev-1)*FLOWERSIZEMULT)

	table.insert(flowers, {gridx=gridx,gridy=gridy,sourcegridx=sourcegridx,sourcegridy=sourcegridy,sourceox=sourceox,sourceoy=sourceoy,genes=genes,flowersize=flowersize,pointed=false,pollinated=false,bee_action=false,bee_counter=0,num_fruits=0,age=0,state='fruit',othergenes=nil,winning_dude=false,g_match=nil} )

	transform_flower_grid_to_screen(newflowerid)

	return( newflowerid )
end

function killflower(i)
	local x = flowers[i].gridx
	local y = flowers[i].gridy
	assert( x <= #flowergrid --[[, "x="..x.." #flowergrid="..#flowergrid ]])
	assert( y <= #flowergrid[x] --[[, "y="..x.." #flowergrid[x]="..#(flowergrid[x]) ]])

	assert(flowergrid[x][y] ~= -1 and flowergrid[x][y] ~= nil )
	flowergrid[x][y] = -1
	flowergridfilled = flowergridfilled - 1
	assert( flowergridfilled >= 0 )
end

function transform_flower_grid_to_screen(flowerid)
	local x = flowers[flowerid].gridx
	local y = flowers[flowerid].gridy

	local tx = x*gridfuncx + gridoffx
	local ty = y*gridfuncy+gridoffy+(gridfuncy/2)*(x%2)

	flowers[flowerid].x = tx
	flowers[flowerid].y = ty

	if flowers[flowerid].sourcegridx ~= nil then
		x = flowers[flowerid].sourcegridx
		y = flowers[flowerid].sourcegridy

		tx = x*gridfuncx + gridoffx
		ty = y*gridfuncy+gridoffy+(gridfuncy/2)*(x%2)

		flowers[flowerid].sourcex = tx
		flowers[flowerid].sourcey = ty
	end
end

function transform_all_flowers_grid_to_screen()
	for i=1,#flowers do
		transform_flower_grid_to_screen(i)
	end
end
