-- Biodiversity by
-- dave@beermex.com

require( 'gene' )
require( 'audio_handler' )
require( 'levels' )
require( 'flowers' )
require( 'TSerial' )

gamestate = 'menu'
currentlevel = nil
graphicsinitialized=false

menu_ITEM = 1
fmenu = 0

highscore_data = {}
highscore_biopeak = 0
highscore_biototal = 0
highscore_level_t = 99*60+99

SCREENMODES = { {width=1024,height=600}, {width=512,height=300}, {width=1366,height=768}, {width=1920,height=1080}, {width=800,height=600}, {width=1024,height=768} }
FULLSCREENMODES = {}
SCREENMODE = 1

function menu_update(dt)
	-- stop any of the looping or long sound effects, except the background birds --

	if not fewbeesound:isStopped() then
		fewbeesound:stop()
	end

	if not manybeesound:isStopped() then
		manybeesound:stop()
	end

	bee1sound_player:stopAll()
	sirenwhistle_player:stopAll()
	fruitgrow_player:stopAll()
end

function go_menu()
	gamestate = 'menu'
end

function menu_keypressed(key)
	if key == 'up' and menu_ITEM > 1 then menu_ITEM = menu_ITEM - 1 end
	if key == 'down' and menu_ITEM < #levels_data then menu_ITEM = menu_ITEM + 1 end
	if key == 'return' or key == 'kpenter' or key == ' ' or key == 'right' then
		go_game(menu_ITEM)
	end
	if key == 'escape' then love.event.push('quit') end
	if menu_ITEM > fmenu + 7 then fmenu = menu_ITEM - 7 end
	if menu_ITEM < fmenu + 3 then fmenu = menu_ITEM - 3 end
	if fmenu + 9 > #levels_data then fmenu = #levels_data-9 end
	if fmenu < 0 then fmenu = 0 end
end


function go_game(level)

	-- this simulation is deterministic
	-- but, it is simple.  and so is randomseed.
	math.randomseed( level )

	flowers_setup()
	current_level = levels_data[level]
	level_goal_genes = current_level.level_goal_genes

	gamestate = 'gamesplash'
	levelstate = ''
	score_text=''
	highscore_text=''
	highscore=false
	MUTATION_ENABLED=true
	bees_mutation_rate=0
	flowers_mutation_rate=0

	biodiversity = { bio_by_gene = {}, total = 0, peak = 0 }

	FEWBEESOUNDVOLUME=0.75
	fewbeesound:setVolume(FEWBEESOUNDVOLUME)
	fewbeesound:play()
	manybeesound:setVolume(0.01)
	manybeesound:play()

	bees_genes = nil -- i.e. bees not laden with pollen yet
	seconds_counter = 0
	entropy_counter = 0
	global_t = 0
	level_t = 0
	level_randomtick_t = 0
	pointed_flower = nil

	setup = current_level.setup()

	transform_all_flowers_grid_to_screen()

	for i=1,#bees do
	    bees[i] = {active=true,vect_x=0,vect_y=0,x=0,y=0,key_x=0,key_y=0,target_x=0,target_y=0,current_x=0,current_y=0,scale=1,t=math.random()*twopi*#bees,red=255,green=255,blue=255}
	end
	bees_t = 0

	mouse_x = 0
	mouse_y = 0
	target_x = 0
	target_y = 0

	ctrl_is_down=false

	cursor_hidden = true
end

function gamesplash_keypressed( key )
    gamestate = 'game'
end

function gamesplash_update()
end

function gamesplash_draw()
	game_draw()

	love.graphics.setFont(menu_FONT)

	love.graphics.setBlendMode('alpha')
	love.graphics.setColor(0,0,0,128)
	love.graphics.rectangle('fill', 1, 1, screenwidth-1,screenheight-1)
	local text = current_level.splash
	local fw = menu_FONT:getWidth(text)
	local z = menuFheight/12
	local yc = (screenheight-menuFheight*2)/2
	love.graphics.setColor(0,0,0)
	love.graphics.print(text, (screenwidth-fw)/2+z, yc+z)
	love.graphics.setColor(255,255,255,192)
	love.graphics.print(text, (screenwidth-fw)/2, yc)

	yc = yc + menuFheight*2
	local text = 'press any key'
	local fw = menu_FONT:getWidth(text)
	local z = menuFheight/12
	love.graphics.setColor(0,0,0)
	love.graphics.print(text, (screenwidth-fw)/2+z, yc+z)
	love.graphics.setColor(255,255,255,192)
	love.graphics.print(text, (screenwidth-fw)/2, yc)
end

function menu_draw()

	love.graphics.setColor( 64, 255, 64 )
	love.graphics.setFont(menu_FONT)
	love.graphics.print( 'up/down/enter to select level. see readme for more keys', 0, 0 )

	local y = (screenheight - 9 * menuFheight)/2
	local maxwidth = 0
	for i=1,#levels_data do
		local w = menu_FONT:getWidth(levels_data[i].caption)
		if w > maxwidth then maxwidth = w end
	end
	maxwidth = maxwidth + menu_FONT:getWidth( '> '..'  ' )

	local mcount = 9
	local t,b,p,prefix,levelscores
	for i = 1,9 do
		if i+fmenu == menu_ITEM then
			prefix = '> '
			love.graphics.setColor( 255,255, 0 )
		else
			prefix = '  '
			love.graphics.setColor( 128, 128, 128 )
		end
		love.graphics.print( prefix..levels_data[i+fmenu].caption, 0, y )
		local sc = highscore_data[ levels_data[i+fmenu].caption ]
		if sc ~= nil then
			t = sc.highscore_level_t
			t = string.format( "Time: %02d:%02d", math.floor(t/60), t%60 )
			b = 'Bio: '..sc.highscore_biototal
			p = 'Peak: '..sc.highscore_biopeak
		else
			t = 'Time: --:--'
			b = 'Bio: -'
			p = 'Peak: -'
		end
		levelscores = t .. ' ' ..b..' '..p
		love.graphics.print( levelscores, maxwidth, y )
		y = y + menuFheight
	end
end

function quitmenu_draw()

	game_draw()

	love.graphics.setFont(menu_FONT)

	love.graphics.setBlendMode('alpha')
	love.graphics.setColor(0,0,0,128)
	love.graphics.rectangle('fill', 1, 1, screenwidth-1,screenheight-1)
	local text = "Exit to main menu - are you sure?"
	local fw = menu_FONT:getWidth(text)
	local z = menuFheight/12
	love.graphics.setColor(0,0,0)
	love.graphics.print(text, (screenwidth-fw)/2+z, (screenheight-menuFheight)/2+z)
	love.graphics.setColor(255,255,255,192)
	love.graphics.print(text, (screenwidth-fw)/2, (screenheight-menuFheight)/2)
end

function setscreenmode( screenmode )
	if fullscreen then
		setscreensize(FULLSCREENMODES[screenmode].width, FULLSCREENMODES[screenmode].height, fullscreen)
	else
		setscreensize(SCREENMODES[screenmode].width, SCREENMODES[screenmode].height, fullscreen )
	end
end

function setscreensize( x, y, fullscreen )
	screenwidth=x
	screenheight=y
	rectdim = math.min(screenwidth,screenheight)
	global_scale = rectdim/600

	gridoffy = screenheight*0.015
	gridfuncy = (screenheight-gridoffy*4-flowerheight-fruitwidth)/(gridheight-0.5)
	gridoffy = gridoffy*4+flowerwidth/2

	gridoffx = screenwidth*0.1
	gridfuncx = (screenwidth-gridoffx*2-flowerwidth*2-fruitwidth)/(gridwidth-1)
	gridoffx = gridoffx+flowerwidth+fruitwidth

	size_swarm = 5 * global_scale
	bee_size = 24/beeimagesize * global_scale

	if false == love.graphics.setMode( screenwidth, screenheight, fullscreen, true, 0 ) then
		fullscreen=false
		love.graphics.setMode( screenwidth, screenheight, false, true, 0 )
	end

	gamefont = love.graphics.newFont(14*global_scale)
	gamefontheight = gamefont:getHeight()

	menu_FONT = love.graphics.newFont(24*global_scale)
	menuFheight = menu_FONT:getHeight() * 1.25

	if fullscreen then
		MOUSEBUG=MOUSEBUG_FULLSCREEN
	else
		MOUSEBUG=MOUSEBUG_WINDOWED
	end

	-- since grid-to-screen function has (potentially) changed, retranslate all the screen
	-- coordinates of the flowers/fruit
	-- This is really only for speed, to save us having to do the grid transform on EVERY screen draw.
	transform_all_flowers_grid_to_screen()

end

function setmousebug( bug )
	if fullscreen then
		MOUSEBUG_FULLSCREEN=bug
		MOUSEBUG=MOUSEBUG_FULLSCREEN
	else
		MOUSEBUG_WINDOWED=bug
		MOUSEBUG=MOUSEBUG_WINDOWED
	end
end

function love.load()

	love.filesystem.setIdentity( 'Biodiversity_spc' )

	if love.filesystem.exists("highscores.lst") then love.filesystem.load("highscores.lst")() end

--[[
	local pline
	for line in love.filesystem.lines("config.txt") do
		pline = string.find(line, "screenwidth=")
		if pline ~= nil and pline[1] == 0 then
			screenwidth=string.sub(line,pline[1])+0
		end
		pline = string.find(line, "screenheight=")
		if pline ~= nil and pline[1] == 0 then
			screenwidth=string.sub(line,pline[1])+0
		end
		pline = string.find(line, "fullscreen=")
		if pline ~= nil and pline[1] == 0 then
			if string.sub(line,pline[1])+0 then
				fullscreen=true
			else
				fullscreen=false
			end
		end
		pline = string.find(line, "mousebugfullscreen=")
		if pline ~= nil and pline[1] == 0 then
			if string.sub(line,pline[1])+0 then
				MOUSEBUG_FULLSCREEN = true
			else
				MOUSEBUG_FULLSCREEN = false
			end
		end
		pline = string.find(line, "mousebugwindowed=")
		if pline ~= nil and pline[1] == 0 then
			if string.sub(line,pline[1])+0 then
				MOUSEBUG_WINDOWED = true
			else
				MOUSEBUG_WINDOWED = false
			end
		end
	end
]]
	beeimage = love.graphics.newImage("bee.png")
	beeimagesize = beeimage:getHeight()
	pointed_flower=nil

	levelwon = love.audio.newSource("97666__Robinhood76__01720_kenacho_flute_indian.ogg",'streaming')
	levelwon:setLooping(false)

	levellost = love.audio.newSource("12737__Leady__panpipes_distant_echo.ogg",'streaming')
	levellost:setLooping(false)

--	bee1sound = love.audio.newSource("103047__Department64__BEE_buzzin_02.ogg",'streaming')
--	bee1sound:setLooping(true)
	bee1sound_player = audio_handler:new("103047__Department64__BEE_buzzin_02.ogg",true)

	fewbeesound = love.audio.newSource("13533__acclivity__BumbleBee.ogg",'streaming')
	fewbeesound:setLooping(true)

	--manybeesound = love.audio.newSource("93571__pete668__Bees_swarming_around_Flowers.ogg",'streaming')
	manybeesound = love.audio.newSource("95699__dobroide__20100423.rosmary.shrub.ogg",'streaming')
	manybeesound:setLooping(true)

	birdsong = love.audio.newSource("70150__Benboncan__Birdsong.ogg",'streaming')
	birdsong:setLooping(true)

--	sirenwhistle = love.audio.newSource("88424__FreqMan__Siren_Whistle_01.ogg", 'streaming')
	sirenwhistle_player = audio_handler:new( "88424__FreqMan__Siren_Whistle_01.ogg" )
--	fruitgrow = love.audio.newSource("54883__lematt__gonflage_ballon1.ogg",'streaming')
--	fruitgrow_player = audio_handler:new("71911__cmusounddesign__je_rubberman.ogg")
	fruitgrow_player = audio_handler:new("balloon_mix.ogg")

--	fruitdrop = love.audio.newSource("13289__schademans__pipe8.ogg",'streaming')
	fruitdrop_player = audio_handler:new("13289__schademans__pipe8.ogg")

--	fruitdone_player = audio_handler:new("54012__Mart1001__Balloon_pop.ogg")
	fruitdone_player = audio_handler:new("33369__herbertboland__mouthpop.ogg")

--	growing = love.audio.newSource("71911__cmusounddesign__je_rubberman.ogg", 'streaming')
--	growing_player = audio_handler:new('')--"71911__cmusounddesign__je_rubberman.ogg")

--	pollen_collected = love.audio.newSource("24628__anamorphosis__GMB_Kantilan_8.ogg", 'streaming')
	pollen_collected_player = audio_handler:new("24628__anamorphosis__GMB_Kantilan_8.ogg")

	birdsong:setVolume(0.25)
	birdsong:play()

	MAX_NUM_FRUITS = 5
	BEE_COUNTER_SCALE = 1/3 -- i.e. 3 seconds required to pollinate and to collect pollen
	FLOWER_LIVING_SECONDS = 1/40 --i.e. 40 seconds to stay alive with no bee action
	FLOWER_GROWING_SECONDS = 1/10 -- i.e. 10 seconds to just grow
	POLLINATED_DELAY_SECONDS = 1/3 -- i.e. 3 seconds to wait before pollinated flowers start to bear fruit


	FLOWERSIZEMIN=5
	FLOWERSIZEMULT=5
	LEAFMULT=2.5
	flowerwidth = (FLOWERSIZEMIN+FLOWERSIZEMULT)*2 -- petals are of size flowersizemin+(flowersizemult*1)
	FLOWERHEIGHTMIN=30
	FLOWERHEIGHTMULT=30
	flowerheight = --[[FLOWERHEIGHTMIN+FLOWERHEIGHTMULT+FLOWERSIZEMIN+FLOWERSIZEMULT]]70
	fruitwidth = 10

	pi = math.pi
	twopi = pi*2
	bee_delay_t = 0
	bee_interp_factor = 4 -- i.e. 4 times persecond we capture a new bee position from the mouse

--[[
	screenwidth=1024
	screenheight= 600
	fullscreen=true
--]]

	graphicsinitialized=false
end

function love.update(dt)

	if gamestate == 'menu' then
		menu_update(dt)
	elseif gamestate == 'game' then
		game_update(dt)
	elseif gamestate == 'gamesplash' then
		gamesplash_update(dt)
	elseif gamestate == 'quitmenu' then
		quitmenu_update(dt)
	end
end

function quitmenu_update(dt)
	fruitgrow_player:pauseAll()
	bee1sound_player:pauseAll()
end

function game_update(dt)

	local flowers = flowers
	local bees = bees
	local dt = dt
	local bee_size = bee_size
	local twopi = twopi

	local mathcos = math.cos
	local mathsin = math.sin
	local mathmax = math.max

	local newhidden

	if not love.keyboard.isDown('lctrl') and not love.keyboard.isDown('rctrl') then
		newhidden = true
	else
		newhidden = false
	end

	mouse_x = love.mouse.getX()
	mouse_y = love.mouse.getY()
	if newhidden ~= cursor_hidden then
		if newhidden then
			love.mouse.setVisible(false)
			if MOUSEBUG then
				love.mouse.setPosition(screenwidth/2,screenheight/2)
			end
		else
			love.mouse.setVisible(true)
			love.mouse.setPosition( mouse_x, mouse_y )
		end
	end
	cursor_hidden = newhidden

	if cursor_hidden then
		target_x = mouse_x
		target_y = mouse_y
--		target_x = love.mouse.getX()
--		target_y = love.mouse.getY()
		assert( target_x ~= nil )
		assert( target_y ~= nil )
	end


	if levelstate == '' then
		level_t = level_t + dt
		level_randomtick_t = level_randomtick_t + dt
		-- every thirty seconds, we generate a random number. just to keep things moving along
		if level_randomtick_t >= 30 then
			level_randomtick_t = 0
			local throwawayarandomnumber = math.random(2)
		end
	end

	global_t = global_t + dt
	seconds_counter = seconds_counter + dt
	if seconds_counter > 1 then seconds_counter = seconds_counter - 1 end
	entropy_counter = entropy_counter + dt
	if entropy_counter > 10 then entropy_counter = entropy_counter - 1; math.random(); end

	bee_delay_t = bee_delay_t + dt*bee_interp_factor
	if bee_delay_t > 1 then
	    bee_delay_t = bee_delay_t - 1
		for i=#bees,1,-1 do
		    if bees[i].active then
				bees[i].key_x = bees[i].target_x
				bees[i].key_y = bees[i].target_y
				if i~=1 then
					bees[i].target_x = bees[i-1].target_x
					bees[i].target_y = bees[i-1].target_y
				end
			elseif(bees[i].current_x < 0 or bees[i].current_y<0) then
				table.remove(bees,i)
			end
		end
		if bees[1].active then
			bees[1].target_x = target_x
			bees[1].target_y = target_y
		end
	end


	bees_t = bees_t + dt*twopi/3 -- one full rotation per three seconds
	for i = 1,#bees do

			local b = bees[i].t + bees_t
			bees[i].scale = bee_size * ( 1 + (0.002 * mathmax(#bees,20))*mathsin(3+b*1.789) )

--		bees[i].key_x = ((bees[i].key_x*140)+(bees[i].target_x))/(141)
--		bees[i].key_y = ((bees[i].key_y*140)+(bees[i].target_y))/(141)

blend_x = (bees[i].target_x * bee_delay_t)+(bees[i].key_x * (1-bee_delay_t))
blend_y = (bees[i].target_y * bee_delay_t)+(bees[i].key_y * (1-bee_delay_t))

----[[
bees[i].current_x = (bees[i].current_x*(1-dt*2))+(blend_x*dt*2)
bees[i].current_y = (bees[i].current_y*(1-dt*2))+(blend_y*dt*2)
----]]

--[[
--sigh
bees[i].current_x = blend_x
bees[i].current_y = blend_y
--]]

--[[
vect_x = (blend_x - bees[i].current_x)
vect_y = (blend_y - bees[i].current_y)
vect_chg_x = vect_x - bees[i].vect_x
vect_chg_y = vect_y - bees[i].vect_y
bees[i].vect_x = bees[i].vect_x + vect_chg_x * dt
bees[i].vect_y = bees[i].vect_y + vect_chg_y * dt
bees[i].current_x = bees[i].current_x + bees[i].vect_x * dt
bees[i].current_y = bees[i].current_y + bees[i].vect_y * dt
--]]

bees[i].x = bees[i].current_x + (mathsin(b)*mathsin(1+b*1.256)*mathcos(2+b*0.456)*size_swarm*(3+#bees/8))
bees[i].y = bees[i].current_y + (mathcos(b*1.01)*mathsin(1+b*1.156)*mathcos(2+b*0.356)*size_swarm*(3+#bees/8))
--bees[i].x = bees[i].current_x
--bees[i].y = bees[i].current_y

	end


	-- lets see which flower the mouse is pointing at
	pointed_flower=nil

	-- process all of the flowers. yeah, this should be in the flowers.lua.  whateva.
	local total_bee_actions=0
	local max_bee_counter=0
	local total_bee_counter=0
	local do_calculate_biodiversity = false
	for p=1,#flowers do
--		if flowers[p].state == 'fruit' and flowers[p].age == 0 then fruitgrow_player:play(p); end

		if flowers[p].state == 'fruit' then
			flowers[p].bee_counter=0
			flowers[p].age = flowers[p].age + dt*.2 -- 5 seconds for fruit to germinate
			if flowers[p].age >= 1 then
				if levelstate == 'won' then
					flowers[p].age = 1
				else
					flowers[p].state = 'growing'
					flowers[p].age = flowers[p].age - 1
					fruitdone_player:play(p,1+0.35*mathsin(global_t+p))
--					fruitgrow_player:stop(p)
--					growing_player:play(p,1+0.05*mathsin(global_t+1))
					do_calculate_biodiversity = true
				end
			end
		elseif flowers[p].state == 'growing' then
			flowers[p].bee_counter=0
			flowers[p].age = flowers[p].age + dt* FLOWER_GROWING_SECONDS -- 10 seconds to grow stem+bud
			if flowers[p].age >= 1 then
				flowers[p].state = 'flowergrowing'
				flowers[p].age = flowers[p].age - 1
				do_calculate_biodiversity = true
			end
		elseif flowers[p].state == 'flowergrowing' then
			flowers[p].bee_counter=0
			flowers[p].age = flowers[p].age + dt*.1 -- 10 seconds to grow flower
			if flowers[p].age >= 1 then flowers[p].state = 'living'; flowers[p].age = flowers[p].age - 1; end
		elseif flowers[p].state == 'living' then
			flowers[p].age = flowers[p].age + dt* FLOWER_LIVING_SECONDS
			if flowers[p].age >= 1 then
				if levelstate == 'won' then
					flowers[p].age = 1
				else
					flowers[p].state = 'dying'
--[[
					flowers[p].pollinated = false
					flowers[p].othergenes = nil
					flowers[p].bee_counter=0
]]
					flowers[p].age = flowers[p].age - 1
				end
			end
		elseif flowers[p].state == 'pollinated' then
			flowers[p].age = flowers[p].age + dt* POLLINATED_DELAY_SECONDS
			if flowers[p].age >= 1 then
				if levelstate == 'won' then
					flowers[p].age = 1
				else
					flowers[p].state = 'fruitgrowing'
					flowers[p].bee_counter=0
					flowers[p].age = flowers[p].age - 1
					fruitgrow_player:play(p,1+0.15*mathsin(global_t+p))
				end
			end
		elseif flowers[p].state == 'fruitgrowing' --[[and levelstate ~= 'won']] then
			flowers[p].bee_counter=0
			flowers[p].age = flowers[p].age + dt*.2 -- 5 seconds to grow fruit after pollination
			if flowers[p].age >= 1 then
				flowers[p].state = 'living'
				flowers[p].age = flowers[p].age - 1
				flowers[p].num_fruits = flowers[p].num_fruits + 1

				newflowerid = addflower(nil,nil,flowers[p].gridx, flowers[p].gridy,-10,-20,flowers[p].genes,flowers[p].othergenes,flowers[p].mutationrate)
				if newflowerid ~= nil then
					transform_flower_grid_to_screen(newflowerid)
					fruitdrop_player:play(p,1+0.1*mathsin(global_t+p))
				end
				fruitgrow_player:stop(p)
				flowers[p].pollinated = false
				flowers[p].othergenes = nil
				flowers[p].mutationrate = 0
				if flowers[p].num_fruits == MAX_NUM_FRUITS then
					flowers[p].state = 'dying'
				end
			end
		elseif flowers[p].state == 'dying' then
			flowers[p].age = flowers[p].age + dt*.2 -- 5 seconds to die
			if flowers[p].age >= 1 then
				killflower(p)
				flowers[p].state = 'dead'
				flowers[p].age = flowers[p].age - 1
				flowers[p].pollinated=false
				flowers[p].bee_action=false
				flowers[p].othergenes=nil
				flowers[p].mutationrate=0
				flowers[p].bee_counter=0
				do_calculate_biodiversity = true
			end
		end

		local anybeeaction=false

		if( levelstate ~= '' or ( flowers[p].state == 'living' or flowers[p].state == 'dying' ) and flowers[p].pollinated==false ) then

			if levelstate == '' and flowers[p].bee_counter < 0 then
				flowers[p].pointed=false
				flowers[p].bee_action=false
				flowers[p].bee_counter = flowers[p].bee_counter + dt
				if flowers[p].bee_counter >= 0 then flowers[p].bee_counter = 0 end
			else

				-- user pointer first
				local dx=mouse_x-flowers[p].x
				local dy=mouse_y-flowers[p].y
				local fsize = flowers[p].flowersize*2*global_scale

				-- simple bounding box rejection first, before we go for the circle math
				if dx >= -fsize and dx <= fsize and dy >= -fsize and dy <= fsize and math.sqrt(dx*dx+dy*dy)<fsize then
					flowers[p].pointed=true
					--[[
					flowers[p].pointed_x = mouse_x
					flowers[p].pointed_x1 = flowers[p].x-fsize
					flowers[p].pointed_x2 = flowers[p].x+fsize
					flowers[p].pointed_y = mouse_y
					flowers[p].pointed_y1 = flowers[p].y-fsize
					flowers[p].pointed_y2 = flowers[p].y+fsize
					--]]
					pointed_flower=p
				else
					flowers[p].pointed=false
				end

				bee_before=flowers[p].bee_action
				flowers[p].bee_action=false

				-- then bees
				if levelstate == '' then
					for i=1,#bees do
						if bees[i].active then

							local dx=bees[i].x-flowers[p].x
							local dy=bees[i].y-flowers[p].y
							local fsize = flowers[p].flowersize*2

							-- simple bounding box rejection first, before we go for the circle math
							if dx >= -fsize and dx <= fsize and dy >= -fsize and dy <= fsize and math.sqrt(dx*dx+dy*dy) then
								flowers[p].bee_action=true

								flowers[p].bee_counter = flowers[p].bee_counter + dt * BEE_COUNTER_SCALE
								if flowers[p].bee_counter >= 1 then
									flowers[p].bee_action=false
									flowers[p].bee_counter = 0
									if bees_genes==nil then
										assert( flowers[p].genes ~= nil )
										bees_genes = flowers[p].genes
										pollen_collected_player:play(nil,1-0.25*mathsin(global_t))
										--cannot collect any more, need to deposit
										-- record the mutation rate at this exact point in time
										bees_mutation_rate = flowers_mutation_rate
										flowers_mutation_rate = 0
										for f=1,#flowers do
											flowers[f].bee_counter=0
											flowers[f].bee_action=false
										end
										flowers[p].bee_counter = -2 -- 2 seconds until next pollen from THIS flower
									else
										assert( flowers[p].pollinated == false )
										assert( flowers[p].othergenes == nil )
										flowers[p].pollinated = true -- redundant, see flag below
										flowers[p].othergenes = bees_genes
										flowers[p].mutationrate = bees_mutation_rate
										sirenwhistle_player:play(p,1+0.1*mathsin(global_t+p))
										flowers[p].age = 0
										flowers[p].state = 'pollinated'
										bees_genes=nil
										bees_mutation_rate = 0
										for f=1,#flowers do
											flowers[f].bee_counter=0
											flowers[f].bee_action=false
										end

									end
								end

								--break  -- you could uncomment this, but then there's no advantage to having lots of bees
							end
						end
					end
				end
			end

			if bee_before~=flowers[p].bee_action then
				if bee_before then
					bee1sound_player:stop(p)
				else
					bee1sound_player:play(p)
					anybeeaction = true
				end
			end

		else

			flowers[p].bee_action=false
			flowers[p].pointed=false
			bee1sound_player:stop(p)

		end

		if anybeeaction then
			fewbeesound:setVolume(0.25)
		else
			fewbeesound:setVolume(FEWBEESOUNDVOLUME)
		end

		if flowers[p].bee_counter > 0 then
			--WORK OUT A MUTATION RATE, BASED ON THE NUMBER AND AMOUNTS OF BEE_COUNTER FROM ALL FLOWERS!
			--STORE THIS ON THE BEES
			total_bee_actions = total_bee_actions + 1
			if flowers[p].bee_counter > max_bee_counter then max_bee_counter = flowers[p].bee_counter end
			total_bee_counter = total_bee_counter + flowers[p].bee_counter
		end

	end

	-- work out the mutation rate from the above
	if total_bee_actions > 1 then
		if MUTATION_ENABLED then
			-- 25% mutation rate per entire flower in addition to the one that gets the most pollen
			-- max 100% obviously
			-- total_bee_counter is between max_bee_counter and total_bee_actions (since each bee_counter counts up to 1 from 0)
			-- so, total_bee_counter - max_bee_counter is the sum of the amount of bee_count across all the OTHER flowers, excluding
			-- the one that has the max bee attention.  for one flower, this between 0 and 1.  therefore, multiply by 0.25. simples.
			flowers_mutation_rate = 0.25 * ( total_bee_counter - max_bee_counter )
			if flowers_mutation_rate > 1 then flowers_mutation_rate = 1 end
		else
			flowers_mutation_rate = 0
		end
	end

	if do_calculate_biodiversity then
		biodiversity = calculate_biodiversity()
	end

end

function love.keypressed(key)

	if key=='v' then
		SCREENMODE = SCREENMODE+1
		if fullscreen then
			if SCREENMODE > #FULLSCREENMODES then SCREENMODE = 1 end
		else
			if SCREENMODE > #SCREENMODES then SCREENMODE = 1 end
		end
		setscreenmode(SCREENMODE)
	elseif key=='f' then
		fullscreen = not fullscreen
		if #FULLSCREENMODES == 0 and fullscreen == true then fullscreen = false end

		if fullscreen then
			if SCREENMODE > #FULLSCREENMODES then SCREENMODE = 1 end
		else
			if SCREENMODE > #SCREENMODES then SCREENMODE = 1 end
		end
		setscreenmode(SCREENMODE)
	elseif key =='m' then
		setmousebug(not MOUSEBUG)
	else

		if gamestate == 'menu' then
			menu_keypressed(key)
		elseif gamestate == 'game' then
			game_keypressed(key)
		elseif gamestate == 'gamesplash' then
			gamesplash_keypressed(key)
		elseif gamestate == 'quitmenu' then
			quitmenu_keypressed(key)
		end

	end
end

function quitmenu_keypressed(key)
	if key == 'y' or key == 'Y' then
		if current_level.caption == 'Testing Mode' then
			if levelstate ~= 'won' then
				level_won()
				gamestate = 'game'
			else
				go_menu()
			end
		else
			go_menu()
		end
	elseif key == 'n' or key == 'N' or key == 'escape' then
		gamestate = 'game'
		fruitgrow_player:resumeAll()
		bee1sound_player:resumeAll()
	end
end

function game_keypressed(key)

	if key == 'escape' then
		if levelstate~='' then
			gamestate='menu'
		else
			gamestate='quitmenu'
		end
		return
	end

--[[
	if key == 'f' then
		addflower()
	end
]]

--[[
]]
end

function set_num_bees( new_num_bees )

	new_num_bees = math.floor(new_num_bees)
	if new_num_bees > 30 then new_num_bees = 30 end
	if new_num_bees < 1 then new_num_bees = 1 end

	local num_active_bees = 0
	for i=1,#bees do
	    if bees[i].active then num_active_bees = num_active_bees + 1 end
	end

	manybeesound:setVolume((new_num_bees-1)/15)

	if new_num_bees > num_active_bees then
 		while #bees < new_num_bees do
	        table.insert(bees,{} )
		end
		for i=1, new_num_bees do
			if bees[i].active == false then
				bees[i].active=true
			elseif bees[i].active == nil then
				local random_color = math.random(255)
				bees[i] = {active=true,vect_x=0,vect_y=0,x=-50,y=-50,target_x=-50,target_y=-50,key_x=-50,key_y=-50,current_x=-50,current_y=-50,scale=1,t=math.random()*twopi*new_num_bees,red=255-random_color/4,green=255-random_color/4,blue=255-random_color }
			end
		end
	end

	if new_num_bees < num_active_bees then
	    for	i=new_num_bees+1,num_active_bees do
--			assert( bees[i].active )
			bees[i].active=false
			bees[i].key_x=-50
			bees[i].key_y=-50
			bees[i].target_x=-50
			bees[i].target_y=-50
		end
	end

	local final_active_bees = 0
	for i=1,#bees do
	    if bees[i].active == true then final_active_bees = final_active_bees + 1 end
	end
	assert( final_active_bees == new_num_bees, "final_active_bees = "..final_active_bees.."  new_num_bees = "..new_num_bees )
end

function printgenes( genes, x, y, r, g, b, direction )

	love.graphics.setColor(r, g, b)
	love.graphics.setLineWidth(1)

	love.graphics.setFont(gamefont)

	local Fh = gamefontheight * 1.25
	assert( Fh > 2 )

	local maxw = 0
	local maxbw = 0
	local maxbb = 0
	local padding = 16
	local numgenes = 0
	local blocksize = 8*global_scale
	local bitstring
	local bit
	local fillmode
	local xp

	for pi,py in pairs(genes) do
		local ww = gamefont:getWidth(py.identifier)
		if ww>maxw then maxw=ww end
		aspec = py.allelespec
		local bitblocks = 0
		for j=1,#aspec do
			bitblocks = bitblocks + blocksize * aspec[j]
			bitblocks = bitblocks + blocksize
		end
		bitblocks = bitblocks - blocksize

		if bitblocks > maxbb then maxbb = bitblocks end

		local bw = bitblocks + padding + gamefont:getWidth(' : '..py:getValueAsDisplay())
		if bw>maxbw then maxbw=bw end
		if direction=='up' then y = y - Fh end
		numgenes = numgenes+1
	end
	love.graphics.setBlendMode('alpha')
	love.graphics.setColor(80,80,80,92)
	if x + maxw+padding+maxbw+2 > screenwidth then x = screenwidth - maxw-padding-maxbw-3 end
	if y < 0 then y = 0 end
	if y + numgenes*Fh > screenheight then y = screenheight - numgenes*Fh-1 end
	love.graphics.rectangle( 'fill', x, y, maxw+padding+maxbw+2, numgenes*Fh )
	love.graphics.setColor(255,255,255,64)
	love.graphics.rectangle( 'line', x, y, maxw+padding+maxbw+2, numgenes*Fh )
	for pi,py in pairs(genes) do
		local text = py.identifier
		local text2 =  ' : '..py:getValueAsDisplay()
		love.graphics.setColor(0,0,0,128)
		love.graphics.print( text, x+1, y+1 )
		love.graphics.print( text, x+2, y+2 )
		love.graphics.print( text2, x+1+maxw+padding+maxbb+padding, y+1 )
		love.graphics.print( text2, x+2+maxw+padding+maxbb+padding, y+2 )
		love.graphics.setColor(r, g, b)
		love.graphics.print( text, x, y )
		love.graphics.print( text2, x+maxw+padding+maxbb+padding, y )

		bitstring = py:getValueAsAlleleBitstring()
		aspec = py.allelespec
		xp = x+maxw+padding
		love.graphics.setLineWidth(1)
		for i=1,#aspec do
			for j=1,aspec[i] do
				bit = bitstring[i]:sub(j*2-1,j*2-1)
				if bit == '1' then fillmode = 'fill' else fillmode = 'line' end
				love.graphics.rectangle( fillmode, xp, y, blocksize-2, blocksize-2 )
				bit = bitstring[i]:sub(j*2,j*2)
				if bit == '1' then fillmode = 'fill' else fillmode = 'line' end
				love.graphics.rectangle( fillmode, xp, y+blocksize, blocksize-2, blocksize-2 )
				xp = xp + blocksize
			end
			xp = xp + blocksize
		end

		y = y + Fh
	end

end

function printbiodiversity()
	love.graphics.setFont(gamefont)
	love.graphics.setColor(255,255,255)
	love.graphics.print( "Biodiversity", 0, 0 )
	local fh = gamefontheight*1.25
	local yc = fh
	for x,y in pairs(biodiversity.bio_by_gene) do
		love.graphics.print( x .. " = " .. y, 0, yc )
		yc = yc + fh
	end
	love.graphics.print( "TOTAL = " .. biodiversity.total, 0, yc )
	yc = yc + fh
	love.graphics.print( "PEAK = " .. biodiversity.peak, 0, yc )
	yc = yc + fh
	if MUTATION_ENABLED then
		local mrate
		if bees_genes ~= nil then mrate = bees_mutation_rate else mrate = flowers_mutation_rate end
		love.graphics.print( "MUTATION RATE = " .. string.format( "%d%%", mrate*100 ), 0, yc )
	else
		assert( bees_mutation_rate == 0 )
		assert( flowers_mutation_rate == 0 )
		love.graphics.print( "(mutation is off)", 0, yc )
	end
end

function initialise_graphics()

	love.graphics.setBackgroundColor(0,0,0)
	fullscreen=true
	FULLSCREENMODES = love.graphics.getModes()
	if FULLSCREENMODES == nil then FULLSCREENMODES = {}; fullscreen=false; end
	SCREENMODE=1
	setscreenmode( SCREENMODE )
	setmousebug(true)
	love.mouse.setVisible(false)
	graphicsinitialized=true

end

function love.draw()

	if not graphicsinitialized then
		initialise_graphics()
	end

	if gamestate == 'menu' then
		menu_draw()
	elseif gamestate == 'game' then
		game_draw()
	elseif gamestate == 'gamesplash' then
		gamesplash_draw()
	elseif gamestate == 'quitmenu' then
		quitmenu_draw()
	end
end

function game_draw()

	local twopi = twopi
	local mathsin = math.sin
	local mathcos = math.cos
	local global_scale = global_scale
	local one_over_global_scale = 1/global_scale
	local global_t = global_t
	local FLOWERHEIGHTMULT = FLOWERHEIGHTMULT

	local brown_r=80
	local brown_g=30
	local brown_b=0
	local budred=80
	local budgreen=255
	local budblue=40

	local flowers = flowers
--[[
	for i=0,gridwidth-1 do
		for j=0,gridheight-1 do
			local tx = i*gridfuncx + gridoffx
			local ty = j*gridfuncy+gridoffy+(gridfuncy/2)*(i%2)
			love.graphics.circle('line',tx,ty,20)
		end
	end
]]

	for i=1,#flowers do
		if flowers[i].pointed_x ~= nil and i==pointed_flower then
			love.graphics.setBlendMode( 'additive' )
			love.graphics.setColor(255,0,0,255)
			love.graphics.rectangle('line',flowers[i].pointed_x1,flowers[i].pointed_y1,flowers[i].pointed_x2-flowers[i].pointed_x1,flowers[i].pointed_y2-flowers[i].pointed_y1)
			love.graphics.circle('fill',flowers[i].pointed_x,flowers[i].pointed_y,5)
			love.graphics.setBlendMode( 'alpha' )
		end
	end

	love.graphics.setBlendMode( 'alpha' )

	love.graphics.setLineStyle( 'smooth' )

	local g_theta = mathsin(global_t*twopi/3)*twopi/64;
	local g_theta_pollinated = global_t*twopi

	local theta = 0
	local ox
	local oy

	for i=1,#flowers do

		--if flowers[i].pointed~=true or math.floor(global_t*4)%2==1 then

			local heightgenevalue = flowers[i].genes.heightgene:getValueAsAlleleValues()[1]
			local heightgene = 1+(heightgenevalue.value/heightgenevalue.maxvalue) -- number between 1 and 2
			local poisonv = flowers[i].genes.poisongene:getValueAsAlleleValues()
			local poison = (poisonv[1].count)/poisonv[1].maxcount
			local numpetalsv = flowers[i].genes.numpetals:getValueAsAlleleValues()[1]
			local numpetals = 2+numpetalsv.value
			local attractgenevalue = flowers[i].genes.attractiveness:getValueAsAlleleValues()[1]
			local attractgene = attractgenevalue.count/attractgenevalue.maxcount -- number between 0 and 1
			assert( poison<=1 )
			assert( heightgene >=1 and heightgene <= 2 )
			local fy = flowers[i].y

			local age = flowers[i].age

			local growscalex, growscaley, flowerscale, printage, fruitalpha, budscale, budrotate
			local flowerstate = flowers[i].state

			if flowerstate == 'fruit' then
				growscalex = 0
				growscaley = 0
				budscale = 0
				flowerscale = 0
				fruitscale = 0.8 + (age*0.2)
				fruitalpha = 1
				printage = false
			elseif flowerstate == 'growing' then
				growscalex = age
				growscaley = growscalex
				budscale = age
				budrotate = 0
				flowerscale = 0
				fruitscale = 1+age*20
				fruitalpha = 1-age*20
				if fruitalpha < 0 then fruitalpha=0 end
				printage = true
			elseif flowerstate == 'flowergrowing' then
				growscalex = 1
				growscaley = 1
				budscale = 1
				budrotate = age
				flowerscale = age
				fruitscale = 0
				printage = true
			elseif flowerstate == 'fruitgrowing' then
				growscalex = 1
				growscaley = 1
				budscale = 1
				budrotate = 0
				flowerscale = 1-age*(1-age)
				fruitscale = 0.8*age*age
				fruitalpha = 1
				printage = true
			elseif flowerstate == 'dying' then
				growscalex = 1+0.5*age
				growscaley = 1-0.85*age
				flowerscale = 1-age
				budscale=0
				fruitscale = 0
				printage = false
			elseif flowerstate == 'dead' then
				growscalex = 1.5
				growscaley = 1-0.85
				flowerscale = 0
				budscale=0
				fruitscale=0
				printage = false
			else -- living or pollinated
				assert( flowerstate == 'living' or flowerstate == 'pollinated' )
				growscalex = 1
				growscaley = 1
				flowerscale = 1
				budscale=0
				fruitscale=0
				printage = (flowerstate == 'pollinated')
			end

			love.graphics.setLineWidth(1)

			love.graphics.translate( flowers[i].x, flowers[i].y ) -- centre of flower
			--love.graphics.circle('line',0,0,45)

			if flowers[i].pointed_x ~= nil and i==pointed_flower then
				love.graphics.setBlendMode( 'additive' )
				love.graphics.setColor(0,255,0,255)
				love.graphics.rectangle('line',flowers[i].pointed_x1-flowers[i].x,flowers[i].pointed_y1-flowers[i].y,flowers[i].pointed_x2-flowers[i].pointed_x1,flowers[i].pointed_y2-flowers[i].pointed_y1)
				love.graphics.circle('fill',flowers[i].pointed_x-flowers[i].x,flowers[i].pointed_y-flowers[i].y,5)
				love.graphics.setBlendMode( 'alpha' )
			end

			love.graphics.scale( global_scale, global_scale )
			love.graphics.translate( FLOWERHEIGHTMULT*0.25*heightgene, FLOWERHEIGHTMULT*heightgene ) -- bottom-most point of stem

			if printage then
				local omage = 1-age
				love.graphics.setColor( 64*omage, 64+191*omage, 64+191*omage )
				love.graphics.scale( 1, 0.25 )
				love.graphics.circle( 'line', 0, 0, 10+20*omage, 10 )
				love.graphics.scale( 1, 4 )
			end

			if flowers[i].bee_counter > 0 then
				if bees_genes==nil then
					love.graphics.setBlendMode( 'alpha' )
					love.graphics.setColor(255,255,255)
					love.graphics.rectangle( 'line', -20, 0, -20, -FLOWERHEIGHTMULT*heightgene )
					love.graphics.setColor(255,255,0)
					love.graphics.rectangle( 'fill', -20, 0, -20, -FLOWERHEIGHTMULT*heightgene*flowers[i].bee_counter )
				else
					love.graphics.setBlendMode( 'alpha' )
					love.graphics.setColor(255,255,255)
					love.graphics.rectangle( 'line', -20, 0, -20, -FLOWERHEIGHTMULT*heightgene )
					love.graphics.setColor(255,192,192)
					love.graphics.rectangle( 'fill', -20, 0, -20, -FLOWERHEIGHTMULT*heightgene*(1-flowers[i].bee_counter) )
				end
			end

			if flowers[i].winning_dude then
				love.graphics.rotate(mathsin(g_theta),mathcos(g_theta))
			end

			if fruitscale ~= 0 then
				if flowers[i].state == 'fruitgrowing' then
					-- here, self is a plant growing a fruit. fruit goes on the plant, halfway up, ish.
					ox=-10
					oy=-10*heightgene
				elseif flowers[i].sourcex ~= nil then
					-- here, the fruit is on a plant. the fruit is plotted at a position interpolated
					-- between target site of the fruit on the ground (sourcex,y) and source site of the fruit
					-- on the plant (sourceox,oy)
					ox=-10
					oy=-10*heightgene
					local fage
					if flowers[i].state == 'fruit' then
						-- fruit is winging it's way across the screen
						fage = (1-age)
						fage = fage * fage * fage
					else
						-- fruit go pop!
						fage = 0
					end

--					fage = 1
--					ox=(flowers[i].sourcex + flowers[i].sourceox/global_scale-flowers[i].x)*(fage)
--					oy=(flowers[i].sourcey + flowers[i].sourceoy/global_scale-flowers[i].y)*(fage)
					ox=(one_over_global_scale*(flowers[i].sourcex - flowers[i].x)+flowers[i].sourceox)*(fage)
					oy=(one_over_global_scale*(flowers[i].sourcey - flowers[i].y)+flowers[i].sourceoy)*(fage)
				else
					ox=0
					oy=0
				end

				love.graphics.setBlendMode( 'alpha' )
				love.graphics.setColor(192,64,0,255*fruitalpha)
				love.graphics.circle('fill',ox,oy,fruitscale*10,32)
				love.graphics.setColor(255,128,0,255*fruitalpha)
				love.graphics.circle('fill',ox+1*fruitscale,oy-1*fruitscale,fruitscale*8,16)
				love.graphics.setColor(192,255,128,255*fruitalpha)
				love.graphics.circle('fill',ox+3*fruitscale,oy-3*fruitscale,fruitscale*4,16)
			end

			if growscale ~= 0 and growscaley ~= 0 then
				love.graphics.scale( growscalex, growscaley )
			end

--			love.graphics.print( ''..flowers[i].state..": "..flowers[i].age, 0, 0 )

			-- draw stem
			local stemred = 60 * (1-poison) + 255*poison
			local stemgreen = 255*(1-poison/2)
			local stemblue=30

			if flowers[i].state=='dying' then
				stemred = stemred*(1-age) + brown_r*age
				stemgreen = stemgreen*(1-age) + brown_g*age
				stemblue = stemblue*(1-age) + brown_b*age
			elseif flowers[i].state=='dead' then
				stemred = brown_r
				stemgreen = brown_g
				stemblue = brown_b
			end

			if growscalex ~= 0 and growscaley ~= 0 then
				love.graphics.setColor(stemred,stemgreen,stemblue)
	--			if flowers[i].pointed and math.floor(global_t*4)%2==1 then love.graphics.setColor(0,0,0) end

				love.graphics.setLineWidth(heightgene*heightgene*growscalex)
				love.graphics.line(0,0,-FLOWERHEIGHTMULT/20*heightgene,-FLOWERHEIGHTMULT*0.5*heightgene,-FLOWERHEIGHTMULT*0.25*heightgene,-FLOWERHEIGHTMULT*heightgene)

				--leaves
				local leafsize = heightgene*LEAFMULT*growscalex*growscaley
				--leaf1
				love.graphics.rotate(twopi/8)
				love.graphics.translate(-leafsize,0)
				love.graphics.scale(1,0.4)
				love.graphics.circle('fill',0,0,leafsize,8)
				love.graphics.scale(1,1/0.4)
				love.graphics.translate(leafsize,0)
				love.graphics.rotate(-twopi/8)
				--leaf2  -  leaf2 is bright green (so is the bud)
				love.graphics.setColor(budred,budgreen,budblue)
				leafsize = leafsize*2
				love.graphics.translate(-FLOWERHEIGHTMULT/20*heightgene,-FLOWERHEIGHTMULT*0.5*heightgene)
				love.graphics.rotate(7*twopi/16)
				love.graphics.translate(-leafsize,0)
				love.graphics.scale(1,0.4)
				love.graphics.circle('fill',0,0,leafsize,8)
				love.graphics.scale(1,1/0.4)
				love.graphics.translate(leafsize,0)
				love.graphics.rotate(-7*twopi/16)
				love.graphics.translate(FLOWERHEIGHTMULT/20*heightgene,FLOWERHEIGHTMULT*0.5*heightgene)
			end

			love.graphics.translate(-FLOWERHEIGHTMULT*heightgene*0.25,-FLOWERHEIGHTMULT*heightgene) -- back to center of flower
			local flowersize = flowers[i].flowersize

			-- bud goes next, so when flower grows, it comes out from the bud
			if budscale ~= 0 then
				local budsize = flowersize*budscale/2
				love.graphics.rotate(twopi/8-twopi*budrotate/4)
				love.graphics.translate(-budsize,0)
				love.graphics.scale(1,0.5)
			    love.graphics.circle('fill',0,0,budsize,8)
				love.graphics.scale(1,2)
				love.graphics.translate(budsize,0)
				love.graphics.rotate(-twopi/4+twopi*budrotate/2)
				love.graphics.translate(budsize,0)
				love.graphics.scale(1,0.5)
			    love.graphics.circle('fill',0,0,budsize,8)
				love.graphics.scale(1,2)
				love.graphics.translate(-budsize,0)
				love.graphics.rotate(twopi/8-twopi*budrotate/4)
			end

			if flowerscale ~= 0 then
				love.graphics.setLineWidth(1.5*growscalex)

				local mcos, msin
				local colorgenevalues = flowers[i].genes.colorgene:getValueAsAlleleValues()
				local petalred = (colorgenevalues[1].value)*240/colorgenevalues[1].maxvalue
				local petalgreen = (colorgenevalues[2].value)*240/colorgenevalues[2].maxvalue
				local petalblue = (colorgenevalues[3].value)*240/colorgenevalues[3].maxvalue

				love.graphics.scale(flowerscale,flowerscale)

				--particles
				if attractgene==1 then
					love.graphics.setBlendMode( 'additive' )
					for j=1,25 do
						local b = ((global_t+i+(j/5))%5)/5
						love.graphics.setColor(255*(1-b),255,255,255*(1-b))
						love.graphics.circle( 'fill', mathsin( (global_t/10)*twopi+j+i)*flowersize*5*b, -b*b*flowersize*5*2, 2 )
					end
					love.graphics.setBlendMode( 'alpha' )
				end

				if flowers[i].state ~= 'dead' then
					if flowers[i].state == 'pollinated' or flowers[i].state == 'fruitgrowing' or flowers[i].winning_dude then
						love.graphics.rotate( g_theta_pollinated )
					elseif flowers[i].state ~= 'living' or flowers[i].age < 0.9 then
						love.graphics.rotate( g_theta )
					end
				end

				-- whitepetals, and poison parts, on bottom
				love.graphics.setColor(255,255-255*poison,0)
	--			if flowers[i].pointed and math.floor(global_t*4)%2==1 then love.graphics.setColor(128,128,128) end
				local petalsize = 1*flowersize*poison*1.35
				local skinniness = 0.1+0.4*(1-poison)
				local unskinniness = 1/skinniness
				for p=1,numpetals do
					theta = (p+0.5) * twopi/numpetals
					mcos = mathcos(theta) * petalsize
					msin = mathsin(theta) * petalsize
					love.graphics.translate( mcos, msin )
					love.graphics.rotate( theta )
					love.graphics.scale( 1, skinniness )
					love.graphics.circle( 'fill', 0, 0, petalsize )
					love.graphics.scale( 1, unskinniness )
					love.graphics.rotate( -theta )
					love.graphics.translate( -mcos, -msin )
				end

				-- main flowers; on top
				local petalsize = 1 * flowersize
				local petalscale = 0.7
				local unpetalscale = 1/petalscale
				for p=1,numpetals do
					theta = p * twopi/numpetals
					mcos = mathcos(theta)*petalsize
					msin = mathsin(theta)*petalsize
					love.graphics.translate( mcos, msin )
					love.graphics.rotate( theta )
					love.graphics.scale( 1, petalscale )
					love.graphics.setColor(0,0,0)
					love.graphics.circle( 'fill', 0, 0, petalsize*1.1 )
					love.graphics.setColor(255,255,255)
					love.graphics.circle( 'fill', 0, 0, petalsize )
					love.graphics.scale( 0.8, 0.8 )
					love.graphics.setColor( petalred, petalgreen, petalblue )
	--				if flowers[i].pointed and math.floor(global_t*4)%2==1 then love.graphics.setColor(255,255,255) end
					love.graphics.circle( 'fill', 0, 0, petalsize )
	--				if flowers[i].pointed and math.floor(global_t*4)%2==1 then love.graphics.setColor(255,255,255) end
					love.graphics.scale( --[[1/0.8, 1/0.8]] 1.25, 1.25 )
					love.graphics.scale( --[[1, 1/0.7]] 1, unpetalscale )
					love.graphics.rotate( -theta )
					love.graphics.translate( -mcos, -msin )
				end

				--center bit
				love.graphics.setColor( 0,0,0 )
				love.graphics.circle('line',0,0,flowersize*0.3)
				love.graphics.setColor( 255,255,64 )
	--			if flowers[i].pointed and math.floor(global_t*4)%2==1 then love.graphics.setColor(255,255,255) end
				love.graphics.circle('fill',0,0,flowersize*0.29)

				love.graphics.scale(1/flowerscale, 1/flowerscale)

				if flowers[i].state ~= 'dead' then
					if flowers[i].state == 'pollinated' or flowers[i].state == 'fruitgrowing' or flowers[i].winning_dude then
						love.graphics.rotate( -g_theta_pollinated )
					elseif flowers[i].state ~= 'living' or flowers[i].age < 0.9 then
						love.graphics.rotate( -g_theta )
					end
				end

			end

			if flowers[i].bee_action then
				c = seconds_counter
				love.graphics.setColor(c*255,c*c*255,c*255)
				love.graphics.setBlendMode( 'additive' )
				love.graphics.circle('line',0,0,flowersize*2*(1+seconds_counter),32)
			end

			if flowers[i].pointed then
				if math.floor(global_t*3)%2==1 then
					love.graphics.setColor(255,255,0,128)
					love.graphics.setBlendMode( 'additive' )
					love.graphics.rotate( -1 * global_t * twopi * 0.25 )
					love.graphics.circle('line',0,0,flowersize*2*1.5,8)
					love.graphics.rotate( global_t * twopi * 0.25 )
				end
--				if flowers[i].pointed_x ~= nil then
--					love.graphics.setColor(255,0,0)
--					love.graphics.rectangle('line',-flowers[i].x+flowers[i].pointed_x1/global_scale,-flowers[i].y+flowers[i].pointed_y1/global_scale,(flowers[i].pointed_x2-flowers[i].pointed_x1)/global_scale,(flowers[i].pointed_y2-flowers[i].pointed_y1)/global_scale)
--					love.graphics.circle('fill',-flowers[i].x+flowers[i].pointed_x/global_scale,-flowers[i].y/global_scale+flowers[i].pointed_y/global_scale,5)
--				end
			end

			love.graphics.setBlendMode( 'alpha' )

			love.graphics.translate( FLOWERHEIGHTMULT*0.25*heightgene, FLOWERHEIGHTMULT*heightgene ) -- translate it back.  we're now back at the bottom of the flower
			if growscale ~= 0 and growscaley ~= 0 then
				love.graphics.scale( 1/growscalex, 1/growscaley )
			end

			if flowers[i].winning_dude then
				love.graphics.rotate(-mathsin(g_theta),-mathcos(g_theta))
			end
			love.graphics.translate( -FLOWERHEIGHTMULT*heightgene*0.25, -FLOWERHEIGHTMULT*heightgene ) -- translate back to center of flower (pre-scale)
			love.graphics.scale( one_over_global_scale, one_over_global_scale ) -- unscale.  now 'completely unscaled'
			love.graphics.translate( -flowers[i].x, -flowers[i].y ) -- translate all the way back to origin.  Now 'completely untranslated'

		--end
	end

	love.graphics.setColor(255,255,255)
	love.graphics.setLineWidth(2)
	local beeimage = beeimage
    for i=1,#bees do
		love.graphics.setBlendMode( 'alpha' )
		love.graphics.setColor(bees[i].red, bees[i].green, bees[i].blue)
		love.graphics.draw(beeimage, bees[i].x-beeimagesize*bees[i].scale*0.5, bees[i].y-beeimagesize*bees[i].scale*0.5, 0, bees[i].scale, bees[i].scale )
    end


	love.graphics.translate( target_x, target_y )
	love.graphics.setLineWidth( 2 * global_scale )

	if bees_genes==nil then
		love.graphics.rotate( global_t * twopi / 6 )
		love.graphics.setBlendMode( 'alpha' )
		love.graphics.setColor( 128,0,128,128 )
		love.graphics.circle( 'line', 0, 0, 12 * global_scale, 6 )
		love.graphics.setBlendMode( 'additive' )
		love.graphics.setColor( 255,128,255 )
		love.graphics.circle( 'line', 0, 0, 10 * global_scale , 6 )
		love.graphics.rotate( -global_t * twopi / 6 )
	else
		printgenes(bees_genes,32*global_scale,-32*global_scale,255,255,255,'up')
		love.graphics.rotate( -global_t * twopi / 6 )
		love.graphics.setBlendMode( 'alpha' )
		love.graphics.setColor( 128,128,0,128 )
		love.graphics.circle( 'line', 0, 0, 12 * global_scale, 3 )
		love.graphics.setBlendMode( 'additive' )
		love.graphics.setColor( 255,255,0 )
		love.graphics.circle( 'line', 0, 0, 10 * global_scale, 3 )
		love.graphics.rotate( global_t * twopi / 6 )
	end

	love.graphics.translate( -target_x, -target_y )

	if pointed_flower ~= nil then
		local f = flowers[pointed_flower]
		local heightgenevalue = f.genes.heightgene:getValueAsAlleleValues()[1]
		local heightgene = 1+(heightgenevalue.value/heightgenevalue.maxvalue) -- number between 1 and 2
		printgenes(f.genes,f.x-32,f.y+40*heightgene,128,255,128,'down')
	end

	for i=1,#flowers do
		if flowers[i].pointed_x ~= nil then
			love.graphics.setColor(0,0,255,255)
			love.graphics.rectangle('line',flowers[i].pointed_x1,flowers[i].pointed_y1,flowers[i].pointed_x2-flowers[i].pointed_x1,flowers[i].pointed_y2-flowers[i].pointed_y1)
			love.graphics.circle('fill',flowers[i].pointed_x,flowers[i].pointed_y,5)
		end
	end

	printbiodiversity()
	printtime()
	if levelstate=='won' or levelstate=='lost' then
		love.graphics.setBlendMode('alpha')
		local text = ''
		local text2 = nil
		local text3 = nil
		if levelstate=='won' then
			love.graphics.setColor((global_t*97)%255, (global_t*83)%255, (global_t*161)%255, 192+63*math.sin(global_t))
			text = 'Target Complete!'
			text2 = score_text
			if highscore_text ~= nil then
				text3 = highscore_text
			end
		else
			love.graphics.setColor(128,128,128,255)
			text = 'LEVEL FAILED!'
		end

		love.graphics.translate(screenwidth/2,screenheight/2)
		local s = 5+3*math.sin(global_t*3)
		love.graphics.scale(s, s)
		local w = menu_FONT:getWidth(text)
		love.graphics.translate(-w/2,-menuFheight/2)
		love.graphics.print( text, 0,0 )
		love.graphics.translate(w/2,menuFheight/2)
		love.graphics.scale(1/s,1/s)

		if text2 ~= nil then
			local w = menu_FONT:getWidth(text2)
			love.graphics.translate(-w/2,-menuFheight/2)
			love.graphics.print( text2, 0, menuFheight*1.5 )
			love.graphics.translate(w/2,menuFheight/2)
		end

		if text3 ~= nil then
			local w = menu_FONT:getWidth(text3)
			love.graphics.translate(-w/2,-menuFheight/2)
			love.graphics.print( text3, 0, -menuFheight*1.5 )
			love.graphics.translate(w/2,menuFheight/2)
		end

		love.graphics.translate(-screenwidth/2, -screenheight/2)
	end
end

function genesort(gene_x, gene_y)
	assert(#gene_x == #gene_y)
	for x,y in ipairs(gene_x) do
		if y.value < gene_y[x].value then return true end
		if y.value > gene_y[x].value then return false end
	end
	return false
end

function calculate_biodiversity( include_growing )
	if include_growing==nil then include_growing=false end

	local uniquegenes={}
	local attractiveness = 0
	local mfloor = math.floor

	local livingflowers = 0
	for i = 1,#flowers do
		if flowers[i].state ~= 'dead' then
			livingflowers = livingflowers + 1
		end
		if flowers[i].state ~= 'dead' and flowers[i].state ~= 'fruit' and (flowers[i].state ~= 'growing' or include_growing) then
			for x,y in pairs(flowers[i].genes) do
				local key=y.identifier
				assert( y ~= nil )
				if uniquegenes[key]==nil then uniquegenes[key]={} end
				v = y:getValueAsAlleleValues()
				assert( v ~= nil )
				assert( #v > 0 )
				uniquegenes[key][(#uniquegenes[key])+1] = v
			end
			if flowers[i].genes.attractiveness ~= nil then
				local a = flowers[i].genes.attractiveness:getValueAsAlleleValues()
				assert( a ~= nil )
				assert( #a > 0 )
			    attractiveness = attractiveness + mfloor(a[1].count/a[1].maxcount)
			end
		end
	end

	if livingflowers == 0 then
		level_failed()
	end

	-- OLD: very attractive flowers attract 2 bees, averagely attractive 1, unattractive zero
	-- NOW: the most attractive flowers attract 1 bee.  Anything else (averagely, or unattractive) attract no extra bees
	set_num_bees( attractiveness )

	local bio_by_gene = {}
	local total = 0
	for x,y in pairs(uniquegenes) do
		assert( #y > 0 )
		for i = 1,#y do assert( y[i] ~= nil ) end
		table.sort(y, genesort)
		local u=0
		local prev = nil
		for a,b in ipairs(y) do
			if prev==nil then
				assert( b ~= nil )
				assert( #b>0 )
				assert(u==0)
				prev = {}
				for c,d in ipairs(b) do prev[c]=d.value end
				u = 0
			else
				local same = true
				assert(prev~=nil)
				assert(#prev>0)
				for c,d in ipairs(prev) do
					assert( d~=nil )
					assert( b[c]~=nil )
					assert( type(b[c].value) == type(d) )
					if d ~= b[c].value then
						same = false
					end
				end
				if not same then
					prev = {}
					for c,d in ipairs(b) do prev[c]=d.value end
					u = u + 1
				end
			end
		end
		bio_by_gene[x] = u
		total = total + u
	end

	test_level_end_genes_condition()

	peak = biodiversity.peak
	if total > biodiversity.peak then peak = total end
	return( { bio_by_gene = bio_by_gene, total = total, peak = peak } )
end

function test_level_end_genes_condition()
	if levelstate == 'won' or levelstate == 'lost' then
		return
	end

	if level_goal_genes ~= nil then
		if level_goal_genes() then
			levelstate = 'won'
			-- calculate final biodiversity, including plants that haven't finished growing
			-- this makes it fair, and makes it include the final goal plant in the scoring, etc
			biodiversity = calculate_biodiversity(true)
		end
	end

	if levelstate == 'won' then
		level_won()
	end
end

function level_won()
	levelstate = 'won'
	levelwon:play()

	score_text = ''
	highscore_text = ''
	highscore = false
	if highscore_data[current_level.caption] == nil then highscore_data[current_level.caption] = {} end

	highscore_level_t = highscore_data[current_level.caption].highscore_level_t
	if highscore_level_t == nil or (current_level.timetarget=='fastest' and level_t < highscore_level_t) or (current_level.timetarget=='longest' and level_t > highscore_level_t) then
		highscore_level_t = level_t
		highscore_data[current_level.caption].highscore_level_t = level_t
		highscore = true
		highscore_text = highscore_text .. 'Time! '
	end
	score_text = score_text .. string.format( "Time: %02d:%02d", math.floor(level_t/60), level_t%60 ) .. "  "

	highscore_biototal = highscore_data[current_level.caption].highscore_biototal
	if highscore_biototal == nil or biodiversity.total > highscore_biototal then
		highscore_biototal = biodiversity.total
		highscore_data[current_level.caption].highscore_biototal = highscore_biototal
		highscore = true
		highscore_text = highscore_text .. 'Final Biodiversity! '
	end
	score_text = score_text .. 'Bio: '.. biodiversity.total .. "  "

	highscore_biopeak = highscore_data[current_level.caption].highscore_biopeak
	if highscore_biopeak == nil or biodiversity.peak > highscore_biopeak then
		highscore_biopeak = biodiversity.peak
		highscore_data[current_level.caption].highscore_biopeak = highscore_biopeak
		highscore = true
		highscore_text = highscore_text .. 'Peak Biodiversity! '
	end
	score_text = score_text .. 'BioPeak: '.. biodiversity.peak

	if highscore then highscore_text = 'HIGHSCORE! '..highscore_text end

	for i=1,#flowers do
		flowers[i].bee_action=false
		flowers[i].bee_counter=0
		flowers[i].othergenes=nil
		flowers[i].mutationrate=0
	end
	bees_genes=nil
	bees_mutation_rate=0
	flowers_mutation_rate=0
	bee1sound_player:stopAll()

	love.filesystem.write("highscores.lst", "highscore_data="..TSerial.pack(highscore_data));

end

function level_failed()
	levelstate = 'lost'
	levellost:play()
end

function printtime()
	love.graphics.setBlendMode('alpha')
	love.graphics.setColor(128,255,255,128)
	love.graphics.setFont(menu_FONT)
	local text = string.format("%02d:%02d", math.floor(level_t/60), level_t%60)
	local w = menu_FONT:getWidth(text)
	love.graphics.print(text, screenwidth-w,0)
end
