-- Biodiversity by
-- dave@beermex.com

Gene={}
mt = {}

-- allelespec is a simple array of bitsizes e.g. {2,2,2} would create a gene with three 2-bit alleles (i.e. each parent provides 3x2=6bits therefore 12 bits in total for the gene)
function Gene:new( identifier, allelespec, valuetype, bitstring )
    if identifier == nil then error( 'Cannot have a blank identifer' ) end
    if allelespec == nil then error( 'Cannot have a blank allelespec' ) end
	if bitstring ~= nil and type(bitstring) ~= 'string' and type(bitstring) ~= 'table' then error( 'Bitstring arg has wrong type' ) end

	if type(bitstring)=='table' then
		bstring=''
		for i=1,#bitstring do
			numstring=''
			value = bitstring[i]
		    assert( type( value ) == 'number' )
			assert( value >= 0 )
		    for j=1,allelespec[i]*2 do
			    if value%2==0 then char='0' else char='1' end
				numstring = char..numstring
				value = math.floor(value*0.5)
			end
			bstring = bstring..numstring
		end
		bitstring = bstring
	end

    local bitcount=0
    for i=1,#allelespec do
        bitcount = bitcount + allelespec[i]*2
	end

    if bitstring == nil then
        bitstring = ''
        for i=1,#allelespec do
            bitstring = bitstring..string.rep('0',allelespec[i]*2)
        end
    else
        if #bitstring ~= bitcount then error( 'new : bitstring has the wrong length!' ) end
    end


    return( setmetatable({ allelespec=allelespec, identifier=identifier, bitcount=bitcount, bitstring=bitstring, valuetype=valuetype, valuedata={} }, mt ))
end

function Gene:copy()
    local newobj = Gene:new(self.identifier, self.allelespec, self.valuetype, self.bitstring)
    return( newobj )
end

-- bitstring is actually just a text string e.g. '00110011010101'
function Gene:setFromBitstring( bitstring )
    if #bitstring ~= self.bitcount then error( 'setFromBitstring : bitstring has the wrong length!' ) end
    self.bitstring = bitstring
	self.valuedata = {}
end

function Gene:getValueAsBitstring()
    return self.bitstring
end

function Gene:getValueAsAlleleBitstring()
	if self.valuedata.allelebitstring == nil then
		local allelebitstring={}
		local aspec = self.allelespec
		local bs = self.bitstring
		local GStart = 1
		for i=1,#aspec do
			local GEnd = GStart + aspec[i]*2 - 1
			allelebitstring[i] = bs:sub(GStart,GEnd)
			GStart = GEnd + 1
		end

		self.valuedata.allelebitstring = allelebitstring
	end
    return self.valuedata.allelebitstring
end

function Gene:getValueAsAlleleValues()
	if self.valuedata.allelevalues == nil then
		local bitstring_values = self:getValueAsAlleleBitstring()
		local retvalues={}
		local value=0
		local bit=0
		local parentAbit=0
		local parentBbit=0
		local char
		local count

		for i=1,#bitstring_values do
			value = 0
			count = 0
			bitstring=''
			m = self.allelespec[i]
			for j=1,#bitstring_values[i],2 do
				char = bitstring_values[i]:sub(j,j)
				if char == '1' then parentAbit = 1 else parentAbit = 0 end
				char = bitstring_values[i]:sub(j+1,j+1)
				if char == '1' then parentBbit = 1 else parentBbit = 0 end

				if parentAbit==1 or parentBbit==1 then bit = 1 else bit = 0 end

				value = value * 2 + bit
				count = count + bit
				if bit==1 then bitstring=bitstring..'1' else bitstring=bitstring..'0' end
			end
			retvalues[i] = {value=value, maxvalue=(2^m)-1, count=count, maxcount=m, bitstring=bitstring}
		end

		assert( retvalues ~= nil )
		self.valuedata.allelevalues = retvalues
	end

	assert( self.valuedata.allelevalues ~= nil )
    return self.valuedata.allelevalues
end

function Gene:getValueAsDisplay()
	if self.valuedata.valueasdisplay == nil then
		local v = self:getValueAsAlleleValues()
		text = ''
		if self.valuetype == 'value' then
			for i=1,#v do
				text = text..v[i].value..'/'..v[i].maxvalue..' '
			end
		elseif self.valuetype == 'count' then
			for i=1,#v do
				text = text..v[i].count..'/'..v[i].maxcount..' '
			end
		else
			assert( false, self.valuetype )
		end
		self.valuedata.valueasdisplay = text
	end
	assert( self.valuedata.valueasdisplay ~= nil )
	return self.valuedata.valueasdisplay
end

-- assumes random number generator has already been seeded.  this function simply pulls out random numbers
function Gene:breed( othergene, mutationrate )
--you can't compare two tables like this.  booh!
    --if pairs(othergene.allelespec) ~= pairs(self.allelespec) then error( 'cannot breed two genes with different allele specs!' ) end

    -- sanity check more than anything else.  this MUST be true.. right?
    if #othergene.bitstring ~= #self.bitstring then error( 'internal error: same allele specs but different bitstring lengths...?!?' ) end

    local mybitstring = self.bitstring
    local yourbitstring = othergene.bitstring
	assert( #mybitstring == #yourbitstring )
	assert( self.valuetype == othergene.valuetype )

    local newbitstring = ''
    local newbit
    local random_number

	local parentAbit=''
	local parentBbit=''

    for i=1,#mybitstring,2 do

		r = math.random(2)-1 -- i.e. 0 or 1
		parentAbit=mybitstring:sub(i+r,i+r)
		r = math.random(2)-1 -- i.e. 0 or 1
		parentBbit=yourbitstring:sub(i+r,i+r)

		if mutationrate ~= nil and mutationrate ~= 0 and math.random() < mutationrate then
			random_number = math.random()
            -- pick a random bit, if mutating
            if random_number>0.5 then parentAbit='0' else parentAbit='1' end
		end
        if mutationrate ~= nil and mutationrate ~= 0 and math.random() < mutationrate then
			random_number = math.random()
            -- pick a random bit, if mutating
            if random_number>0.5 then parentBbit='0' else parentBbit='1' end
		end


        newbitstring = newbitstring..parentAbit..parentBbit
    end

    self.bitstring = newbitstring
	self.valuedata = {}
end

mt.__index = Gene

